param (
    [string]$VCPath = "C:/Program Files (x86)/Microsoft Visual Studio 14.0/VC",
    [Parameter(Mandatory=$false)][ValidateSet("true", "false")][string]$SkipContent="false",
    [Parameter(Mandatory=$false)][ValidateSet("true", "false")][string]$SkipBoost="false",
    [Parameter(Mandatory=$false)][ValidateSet("true", "false")][string]$SkipProtobuf="false",
    [Parameter(Mandatory=$false)][ValidateSet("true", "false")][string]$SkipGTest="false",
    [Parameter(Mandatory=$false)][ValidateSet("true", "false")][string]$UseSubwindows="false"
)

Function Parse-Bool([string]$BoolString){
	try{
        $ParsedBool = $false
        switch($BoolString.ToLower()) {
            "true" { $ParsedBool = $true }
            default { $ParsedBool = $false }
        }

        return [bool]$ParsedBool
    }
	catch{}
}

Function Try-CreateDirectory([string]$DirPath){
	try{
        if(-Not (Test-Path $DirPath))
        {
            New-Item -ItemType directory -Path $DirPath
        }
    }
	catch{}
}

# Invokes a Cmd.exe shell script and updates the environment.
function Invoke-CmdScript {
  param(
    [String] $EnvPath,
    [String] $ScriptName,
    [String] $WorkingDirectory,
    [bool] $UseSubwindows
  )

  Write-Host "Getting MSVC environment from $EnvPath"

  if($UseSubwindows)
  {
    $p = Start-Process "cmd" -ArgumentList "/c","`"$EnvPath`" & $ScriptName" -WorkingDirectory "$WorkingDirectory" -PassThru
    $p.WaitForExit()
  }
  else
  {
    $p = Start-Process "cmd" -ArgumentList "/c","`"$EnvPath`" & $ScriptName" -WorkingDirectory "$WorkingDirectory" -NoNewWindow -PassThru
    $p.WaitForExit()
  }
}

$ScriptDir = (Get-Item "$($PSScriptRoot)").FullName
$VCBinPath = Join-Path -Path $VCPath -ChildPath "bin"
$VCEnvBatPath = Join-Path -Path $VCPath -ChildPath "bin/amd64/vcvars64.bat"

# Parse bool to avoid conflicts with options
$SkipContentBool = Parse-Bool $SkipContent
$SkipBoostBool = Parse-Bool $SkipBoost
$SkipProtobufBool = Parse-Bool $SkipProtobuf
$SkipGTestBool = Parse-Bool $SkipGTest
$UseSubwindowsBool = Parse-Bool $UseSubwindows

Write-Host "Launching CARLA setup..."

# Check if requirements are met
Write-Host "Checking powershell dependencies ..."
$ErrorCount = 0

# Command py -3.6 must work
$IsPythonAvailable = [bool](Get-Command "py" -ea 0)
if(-Not $IsPythonAvailable)
{
    write-error "Could not find py, please install python launcher (and python 3.6) and make sure py -3.6 is available"
    $ErrorCount++
}
else
{ 

    $GetInstalledPython = (Invoke-Expression "py -V")
    if(-Not $GetInstalledPython -contains "Python 3.6")
    {
        write-error "Could not find python 3.6"
        $ErrorCount++
    }
}

# Command cmake must work
$IsCmakeAvailable = [bool](Get-Command "cmake" -ea 0)
if(-Not $IsCmakeAvailable)
{
    write-error "cmake failed, install cmake and add it to the PATH"
    $ErrorCount++
}

# Command git must work
$IsGitAvailable = [bool](Get-Command "git" -ea 0)
if(-Not $IsGitAvailable)
{
    write-error "git failed, install cmake and add it to the PATH"
    $ErrorCount++
}

# Make sure VCPath is correct
$ClPath = Join-Path -Path $VCBinPath -ChildPath "cl.exe"
$ClFound = Test-Path $ClPath
if(-Not $ClFound)
{
    Write-Host "$($ClPath) does not exist"
    $ErrorCount++
}

#Check if 7z is available
$Is7ZPowershellCmdAvailable = [bool](Get-Command "Expand-7zip" -ea 0)
if(-Not $Is7ZPowershellCmdAvailable)
{
    Install-Package 7Zip4Powershell
}

# Exit if requirements are not met
if($ErrorCount -gt 0)
{
    Write-Host "Build failure"
    exit
}
else
{
    Write-Host "Checking powershell dependencies -- Done"
}
# Download content
$ContentFolder = Join-Path -Path $ScriptDir -ChildPath "Unreal/CarlaUE4/Content"
Write-Host "Downloading content"
if((-Not (Test-Path $ContentFolder)) -and (-Not $SkipContentBool))
{
    New-Item -ItemType directory -Path $ContentFolder

    $CarlaContentVersionText = Get-Content (Join-Path -Path $ScriptDir -ChildPath Util/ContentVersions.txt) | Select-String -notmatch "#" 

    $CarlaContentVersionList = $CarlaContentVersionText | Foreach-Object{ 
        if(-Not [string]::IsNullOrEmpty($_))
        {
            $splits = $_ -split ':'

            New-Object PSobject -Property @{
            Version=[System.Version]($splits[0].ToString().Trim())
            Id = $splits[1].ToString().Trim()
            } | Select-Object Version,Id
        }
    }

    $newest = $CarlaContentVersionList[0]
    foreach ($version in $CarlaContentVersionList) {
	    if($version.Version -gt $newest.Version){
		    $newest = $version
	    }
    }

    Set-Location (Join-Path -Path $ScriptDir -ChildPath Util)

    Write-Host $newest

    py -3.6 -m pip install -r requirements.txt
    #py -3.6 download_from_gdrive.py $newest.Id Content.tar.gz

    Expand-7zip Content.tar.gz .
    Expand-7zip Content.tar $ContentFolder

    Remove-Item Content.tar.gz
    Remove-Item Content.tar

    Write-Host "Dowloading content -- Done"

    Set-Location $ScriptDir
}
else
{
    if($SkipContentBool)
    {
        Write-Host "Skipping content download required..."
    }
    else
    {
        Write-Host "Folder Unreal/CarlaUE4/Content already exists, skipping download..."
    }
}

# Build dependencies
$UtilBuildPath = Join-Path -Path $ScriptDir -ChildPath Util/Build
Try-CreateDirectory $UtilBuildPath
Set-Location $UtilBuildPath

if(-Not $SkipBoostBool)
{
    $BoostSourcePath = "https://dl.bintray.com/boostorg/release/1.64.0/source/boost_1_64_0.tar.gz"

    Write-Host "Building boost"
    if(-Not (Test-Path boost-source))
    {
        $WebClient = New-Object System.Net.WebClient
        $WebClient.DownloadFile($BoostSourcePath, (Join-Path -Path $UtilBuildPath -ChildPath "boost_1_64_0.tar.gz"))
        Expand-7zip boost_1_64_0.tar.gz .
        Expand-7zip boost_1_64_0.tar .

        Move-Item boost_1_64_0 boost-source

        Remove-Item boost_1_64_0.tar.gz
        Remove-Item boost_1_64_0.tar
    }
    else
    {
        Write-Host "Folder boost-source already exists, skipping download..."
    }

    Set-Location boost-source
    
    # Compile boost
    Invoke-CmdScript $VCEnvBatPath "bootstrap.bat & b2 clean & b2 variant=release link=static runtime-link=shared threading=multi address-model=64" (Join-Path -Path $UtilBuildPath -ChildPath boost-source) $UseSubwindowsBool

    if(Test-Path ../boost-install)
    {
        Remove-Item ../boost-install -recurse -Force
    }

    Try-CreateDirectory ../boost-install

    Move-Item stage ../boost-install -Force
    Copy-Item boost -Destination ../boost-install -recurse -Force

    Set-Location $UtilBuildPath
    Write-Host "Building boost -- Done" 
}

#protobuf section
if(-Not $SkipProtobufBool)
{
    Write-Host "Building protobuf"
    if(-Not (Test-Path protobuf-source))
    {
        # see https://github.com/google/protobuf/issues/3047 for error with 3.3.0 (missing files) 
        git clone --depth=1 -b v3.3.2 https://github.com/google/protobuf.git protobuf-source
    }
    else
    {
        Write-Host "Folder protobuf-source already exists, skipping git clone..."
    }

    Set-Location protobuf-source/cmake
    Try-CreateDirectory build
    Try-CreateDirectory ../../protobuf-install
    Set-Location build

    Invoke-CmdScript $VCEnvBatPath "cmake -DCMAKE_GENERATOR_PLATFORM=x64 ^ -DCMAKE_BUILD_TYPE=Release ^ -DCMAKE_INSTALL_PREFIX=../../../protobuf-install ^ -Dprotobuf_BUILD_TESTS=OFF ^ -DCMAKE_CXX_FLAGS_RELEASE=/MD ^ -Dprotobuf_MSVC_STATIC_RUNTIME=OFF ^ ../ & cmake --build . --target INSTALL --config Release & exit" (Join-Path -Path $UtilBuildPath -ChildPath protobuf-source/cmake/build/) $UseSubwindowsBool

    Set-Location $UtilBuildPath
    Write-Host "Building protobuf -- Done" 
}

#googletest section
if(-Not $SkipGTestBool)
{
    Write-Host "Building googletest"
    if(-Not (Test-Path googletest-source))
    {
        git clone --depth=1 -b release-1.8.0 https://github.com/google/googletest.git googletest-source
    }
    else
    {
        Write-Host "Folder googletest-source already exists, skipping git clone..."
    }

    Try-CreateDirectory googletest-install

    Set-Location googletest-source/googletest

    Try-CreateDirectory cmake/build
    Set-Location cmake/build

    Invoke-CmdScript $VCEnvBatPath "cmake -DCMAKE_GENERATOR_PLATFORM=x64 ^ -DCMAKE_BUILD_TYPE=Release ^ -DCMAKE_INSTALL_PREFIX=../../../../googletest-install ^ ../.. & cmake --build . --target INSTALL --config Release & exit" (Join-Path -Path $UtilBuildPath -ChildPath googletest-source/googletest/cmake/build/) $UseSubwindowsBool

    Set-Location $UtilBuildPath
    Write-Host "Building googletest -- Done" 
}

# other carla files region

Invoke-CmdScript $VCEnvBatPath "Protoc.bat" (Join-Path -Path $ScriptDir -ChildPath Util/) $UseSubwindowsBool

Write-Host ""
Write-Host "****************"
Write-Host "*** CARLA setup... - Done! ***"
Write-Host "****************"